/**********************************************************************
 * imap4_out.c                                               March 1999
 * Horms                                             horms@vergenet.net
 *
 * Functions to communicate with upstream IMAP4 server
 *
 * Released under the terms of the GNU GPL
 *
 **********************************************************************/

#include "imap4_out.h"
#include "options.h"

/**********************************************************************
 * imap4_authenticate
 * Authenticate user with backend imap4 server
 * pre: in_fd: File descriptor to read server responses from
 *      out_fd: File descriptor to write to server on
 *      pw:     structure with username and passwd
 *      tag:    tag to use when authenticating with back-end server
 *      protocol: protocol structiure for imap4
 *      buf:    buffer to return response from server in
 *      n:      size of buf in bytes
 * post: 1: on success
 *       0: on failure
 *       -1 on error
 **********************************************************************/

int imap4_out_authenticate(
  const int in_fd, 
  const int out_fd, 
  const struct passwd *pw,
  const token_t *tag,
  const protocol_t *protocol,
  unsigned char *buf,
  size_t *n
){
  token_t *ok;
  queue *q;
  char *tag_string=NULL;
  char *read_string=NULL;
  char *greeting_string=NULL;
  int status=-1;

  extern options_t opt;

  if((ok=create_token())==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_authenticate: create_token");
    }
    goto leave;
  }
  assign_token(ok, (PERDITION_USTRING) IMAP4_OK, strlen(IMAP4_OK), -1);

  if((status=imap4_out_response(in_fd, IMAP4_UNTAGED, ok, &q, NULL, NULL))<0){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_authenticate: imap4_out_response");
    }
    goto leave;
  }
  else if(!status){
    status=0;
    goto leave;
  }

  if((read_string=perdition_queue_to_string(q))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "pop3_out_authenticate: perdition_queue_to_string");
    }
    status=-1;
    goto leave;
  }

  if((greeting_string=greeting_str(
    greeting_string, 
    protocol,
    GREETING_ADD_NODENAME
  ))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_authenticate: greeting_str");
    }
    goto leave;
  }

  if((status=strcmp(read_string, greeting_string))==0){
    syslog(LOG_WARNING, "Loop detected, abandoning connection");
    goto leave;
  }

  if((tag_string=token_to_string(tag))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_in_tagged_ok: token_to_string");
    }
    return(-1);
  }

  if(write_str(
    out_fd, 
    NULL_FLAG,
    5, 
    tag_string, 
    " LOGIN ",
    pw->pw_name,
    " ",
    pw->pw_passwd)<0
  ){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_authenticate: imap4_write");
    }
    status=-1;
    goto leave;
  }


  if((status=imap4_out_response(in_fd, tag_string, ok, &q, buf, n))<0){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_authenticate: imap4_out_response 2");
    }
  }
  destroy_queue(q, DESTROY_TOKEN);

  leave:
  str_free(&tag_string);
  str_free(&greeting_string);
  unassign_token(ok);
  destroy_token(&ok);
  return(status);
}
  

/**********************************************************************
 * imap4_out_response
 * Compare a respnse from a server with the desired response
 * pre: in_fd: file descriptor to read from
 *      tag_string: tag expected from server
 *      desired_token: token expected from server
 *      q: resulting queue is stored here
 *      buf: buffer to read server response in to
 *      n: size of buf
 * post: 1 : tag and desired token found
 *       0: tag and desired token not found
 *       -1: on error
 **********************************************************************/

int imap4_out_response(
  const int in_fd, 
  const char *tag_string,
  const token_t *desired_token,
  queue **q,
  unsigned char *buf,
  size_t *n
){
  int status;
  int tagged;
  token_t *t;
  char *server_tag_string;

  extern options_t opt;

  *q=NULL;
  server_tag_string=NULL;
  t=NULL;
  status=0;

  /*
   * tagged set to 0 if an untagged message is expected
   * set to non-zero otherwise
   */
  tagged=strcmp(tag_string, IMAP4_UNTAGED);

  /*Check tag*/
  while(1){
    if((*q=read_line(in_fd, buf, n))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_out_response: read_line");
      }
      return(-1);
    }
  
    if((*q=pop(*q, (perdition_queue_t *)&t))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_out_response: pop");
      }
      return(-1);
    }

    if((server_tag_string=token_to_string(t))==NULL){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_out_response: token_to_string");
      }
      goto leave;
    }
    destroy_token(&t);
  
    if(tagged && !strcmp(server_tag_string, IMAP4_UNTAGED)){
      destroy_queue(*q, DESTROY_TOKEN);
      continue;
    }
  
    if(strcmp(server_tag_string, tag_string)){
      if(opt.debug){
        syslog(LOG_WARNING, "imap4_out_resonse: invalid tag from server");
      }
      goto leave;
    }
  
    break;
  }

  if((*q=pop(*q, (perdition_queue_t *)&t))==NULL){
    if(opt.debug){
      syslog(LOG_WARNING, "imap4_out_response: pop");
    }
    return(-1);
  }
  
  /*Check token*/
  status=token_cmp(desired_token, t);

  leave:
  destroy_token(&t);
  if(status!=1){
    destroy_queue(*q, DESTROY_TOKEN);
    *q=NULL;
  }
  str_free(&server_tag_string);
  return(status);
}
